import bpy
import bmesh
from . PS_LIB import *
from . PS_PROFILER import *

#from time import time

def SWEEP_GEN(U_normal,sweep,close,fuse):
    #t=time()
    # __________________________________________________
    # ______________________________________OP VARIABLES

    unwrap=False
    sweep_ramp = False
    profile_ramp = False
    go_math = False
    dashed = False

    
    # ______________________________________________________
    # ____________________________________________INIT LISTS

    #BASE_PATH = set()
    PATH=set() #duplicate path loop

    OPP_F = []  # opposite faces in the loop
    V_OPP = set()  # EDGE RING
    ER = set()  # EDGE RING
    EL = set()  # LOOP EDGE A+B

    A_LOOP = []  # EDGE LOOP A RAIL
    A_VERTS = set()  # EDGE LOOP A RAIL

    SWEEP = set()  # sweeped geometry
    PROFILE_FACES = set()
    OPP_FACES = []


    if not profile_ramp:
        FP = FaceProfile()
        if type(FP) is str: return ['ERROR', FP] # @@@ ERROR @@@
        Zlen, PzMIN, PzMAX, PROFILE = ZLEN_MIN_MAX(FP[0])
        CV=FP[1]
    else:
        return ['ERROR', 'profile error'] # @@@ ERROR @@@

    # ___________________________________________________
    # ___________________________________________LOAD GEO
    
    bpy.ops.object.editmode_toggle()
    bpy.ops.object.editmode_toggle()

    obj = bpy.context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    PATH_CENTER=set()
    BASE_PATH = set(( f for f in bm.faces if f.select)) #GET FACE LOOP 

    if len(BASE_PATH)==0: return ['ERROR', 'select face loop'] # @@@ ERROR @@@
    if (sweep_ramp or go_math) and  not 'PS_prof_curve_ramp' in bpy.data.node_groups:
        return ['ERROR', '#5 press INIT RAMP NODE button'] # @@@ ERROR @@@  

    for f in BASE_PATH:
        nu_f = f.copy()
        f.select=False
        nu_f.select=True
        PATH.add(nu_f)
        PATH_CENTER.add((
            (float (str (f.calc_center_median().x)[:5])),
            (float (str (f.calc_center_median().y)[:5])),
            (float (str (f.calc_center_median().z)[:5]))
            ))
    
    bpy.ops.mesh.remove_doubles(use_unselected=False)
    bpy.ops.mesh.hide(unselected=True)

   #__________________________________________SELECTION VALIDATION

    LE_CTRL=[]
    
    for f in PATH:
        for v in f.verts:
            LE = len(v.link_edges)
            if LE > 3:
                ROLLBACK(BASE_PATH,PATH)
                return ['ERROR', '#1 wrong face loop selection'] # @@@ ERROR @@@
            else:
                LE_CTRL.append(LE)
    
    link_edge_set=set(LE_CTRL)
    
    if len(PATH)==1:
        for f in PATH:
            if len( f.verts ) != 4:
                return ['ERROR', '#2 wrong face loop selection'] # @@@ ERROR @@@

    elif len(link_edge_set) == 1:
        if link_edge_set.pop() != 3:
            ROLLBACK(BASE_PATH,PATH)
            return ['ERROR', '#3 wrong face loop selection'] # @@@ ERROR @@@
    
    elif LE_CTRL.count(2)!=4:
            ROLLBACK(BASE_PATH,PATH)
            return ['ERROR', '#4 wrong face loop selection'] # @@@ ERROR @@@


    #_____________________________________________________________

# _____________________________________________________
# _____________________________________INITIALIZE LISTS

    ET = set((e for e in bm.edges if e.select))

    bpy.ops.mesh.region_to_loop()

    EB = set((e for e in bm.edges if e.select))

    for f in PATH:  # FIND FIRST AND LAST FACES IN THE LOOP
        n = 0
        for e in f.edges:
            if (e in EB): n += 1
        if (n > 2): OPP_F.append(f)

    if (len(OPP_F) == 2):  # LOOP FACE CASE -> GET EDGE RING
        for f in OPP_F:
            opp_e = set(f.edges) & EB
            alt_v = list(set(f.edges)-opp_e)
            alt_v = alt_v[0].verts
            for e in opp_e:
                if not (e.verts[0] in alt_v or e.verts[1] in alt_v):
                    ER.add(e)
    
    elif (len(OPP_F) == 1):  # SINGLE FACE CASE
        start_edge = Cursor_edge(EB)
        opp_v = start_edge.verts
        for e in OPP_F[0].edges:
            if ((e.verts[0] in opp_v) ^ (e.verts[1] in opp_v)):
                ER.add(e)

    EL = EB-ER  # LOOP EDGE A+B
    ER = ET-EL | ER  # EDGE RING
    
    E_OPP = EB & ER
    #print (len(E_OPP))
    #if len(E_OPP)==0: return ['ERROR', 'wrong face loop selection'] # @@@ ERROR @@@

    for e in E_OPP:
        for v in e.verts:
            V_OPP.add((v.co.x, v.co.y, v.co.z))

    edge_A = Cursor_edge(EL)  # SORTED LOOP START
    
    planar_length = sum([e.calc_length() for e in ER])/len(ER) # to equalize sweep flow
 
    def grow_loop(edge):
        if not edge in A_LOOP:
            A_LOOP.append(edge)  # test
        for v in edge.verts:
            links = v.link_edges
            for link in links:
                if (link in EL and link not in A_LOOP):
                    A_LOOP.append(link)  # test
                    grow_loop(link)

    grow_loop(edge_A)
    #B_LOOP = EL-set(A_LOOP)

    for e in A_LOOP:
        for v in e.verts:
            A_VERTS.add(v)

    for e in ET: e.select = False
 

    ER_len = len(ER)
                    
    SORTED_RING = ER
    r_count=0
    C_mod=1        

# _____________________________________________________
# _____________________________________ SWEEP GEN CODE


    #print ('SORTING',time()-t)
    #t=time()
    
    for e in SORTED_RING:
        r_count+=1
        for v in e.verts:
            if v in A_VERTS:
                V0 = v
            else:
                V1 = v

        PROJ = V1.co-V0.co
        
        L = PROJ.length

        L = L/Zlen  # L:x=Zlen:1

        NU_VERTS = []

        PROJ.normalize()
        
        S = V0.calc_shell_factor()  # SCALE FACTOR ANGLE BASED

        normal = [V0.normal[0], V0.normal[1], V0.normal[2]]

        for v in PROFILE:  # PROJECT PROFILE POINT ALONG NORMAL

            z0 = v[2]
            x0 = v[0]

            if sweep_ramp or go_math: C_mod=C_normal[r_count-1]
           
            X = V0.co.x - (PROJ[0]*z0*L) + (normal[0]* x0*S*U_normal*C_mod) * (planar_length/Zlen)
            Y = V0.co.y - (PROJ[1]*z0*L) + (normal[1]* x0*S*U_normal*C_mod) * (planar_length/Zlen)
            Z = V0.co.z - (PROJ[2]*z0*L) + (normal[2]* x0*S*U_normal*C_mod) * (planar_length/Zlen)

            v = bm.verts.new((X, Y, Z))
            NU_VERTS.append(v)
            SWEEP.add(v)


        f = bm.faces.new(NU_VERTS)
        f.select = True

    #print ('VERTS PROJECTED',time()-t)
    #t=time()


    # ____________ PROFILE FACE GENERATED AND SELECTED --> SWEEP CODE --->

    PROFILE_FACES = set(( f for f in bm.faces if f.select ))

    if (sweep == True and (len(OPP_F) == 2 or len(PATH) > 2)):  # ORDERED FACE SWEEP
        bpy.ops.mesh.select_all(action='DESELECT')

        SWEEP_FACES = set()
        last_edge = len(A_LOOP)-1

        #t=time()
        for i, e in enumerate(A_LOOP):
            
            v1 = e.verts[0].co
            v2 = e.verts[1].co
            f1 = None
            f2 = None
            
            for f in PROFILE_FACES:
                for v in f.verts:
                    if v.co == v1:
                        f1 = f.copy()
                        if close and (v.co.x, v.co.y, v.co.z) in V_OPP:  # check first edge A loop
                            first_face = f.copy()
                            SWEEP_FACES.add(first_face)
                            OPP_FACES.append(first_face)

                    elif v.co == v2:
                        f2 = f.copy()
                        if close and (v.co.x, v.co.y, v.co.z) in V_OPP:  # check first edge B loop
                            first_face = f.copy()
                            SWEEP_FACES.add(first_face)
                            OPP_FACES.append(first_face)
            
            if f1 == None or f2 == None:

                return "WRONG SELECTION"


            
            f1.select = True
            f2.select = True
            bpy.ops.mesh.bridge_edge_loops(type='SINGLE')
            for f in bm.faces:
                if f.select: 
                    SWEEP_FACES.add(f)

            bpy.ops.mesh.select_all(action='DESELECT')
            f1.select = True
            f2.select = True
            bpy.ops.mesh.delete(type='FACE')
        
        #print ('SWEEP',i,time()-t)##############TIMING

        if unwrap and len(E_OPP)==0:
            PROFILE_FACE = PROFILE_FACES.pop()
            
        for f in PROFILE_FACES:
            f.select=True

        bpy.ops.mesh.delete(type='FACE')

        #if len(OPP_F) == 0:
            #for f in PROFILE_FACES: f.select = True
            #bpy.ops.mesh.delete(type='FACE')

        for f in SWEEP_FACES:
            f.select=True   
        
    elif (sweep == True and len(OPP_F) == 1):  # SINGLE FACE SWEEP
        for f in PATH:
            f.select = False
        bpy.ops.mesh.bridge_edge_loops(type='SINGLE')

   
    
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.remove_doubles(use_unselected=False)
 
    bpy.ops.mesh.normals_make_consistent()
    
  
    bpy.ops.mesh.reveal(select=False)

    if fuse:
        bpy.ops.mesh.select_all(action='DESELECT')
        for f in bm.faces:
            if (((float (str (f.calc_center_median().x)[:5])),
                (float (str (f.calc_center_median().y)[:5])),
                (float (str (f.calc_center_median().z)[:5])))) in PATH_CENTER:
                if dashed:
                    if (((float (str (f.calc_center_median().x)[:5])),
                        (float (str (f.calc_center_median().y)[:5])),
                        (float (str (f.calc_center_median().z)[:5])))) in DASHED_BLOCKS: f.select=True
                else:
                    f.select=True

        bpy.ops.mesh.select_linked(delimit={'NORMAL'})
        bpy.ops.mesh.remove_doubles(use_unselected=False)
        bpy.ops.mesh.select_all(action='DESELECT')


        for f in bm.faces:
            if (((float (str (f.calc_center_median().x)[:5])),
                (float (str (f.calc_center_median().y)[:5])),
                (float (str (f.calc_center_median().z)[:5])))) in PATH_CENTER:
                if dashed:
                    if (((float (str (f.calc_center_median().x)[:5])),
                        (float (str (f.calc_center_median().y)[:5])),
                        (float (str (f.calc_center_median().z)[:5])))) in DASHED_BLOCKS: f.select=True
                else:
                    f.select=True

        bpy.ops.mesh.delete(type='FACE')
        bpy.ops.mesh.select_all(action='SELECT')
        #bpy.ops.mesh.remove_doubles(use_unselected=False)
        bpy.ops.mesh.select_all(action='DESELECT')

    bpy.ops.mesh.select_mode(type="FACE")
    bmesh.update_edit_mesh(me) 
    
    return ['OK','SWEEPED!']


def ROLLBACK(b,p):
    bpy.ops.mesh.reveal()
    bpy.ops.mesh.select_all(action='DESELECT')
    for f in p: f.select=True
    bpy.ops.mesh.delete(type='FACE')
    for f in b: f.select=True

